import datetime
from contextlib import suppress
from itertools import chain
from typing import Protocol, final

import attrs
import httpx
from packaging import version
from typing_extensions import TypeAlias, TypedDict


class VersionInfo(TypedDict):

    upload_time: str


VersionNumber: TypeAlias = str
SortedVersionsList: TypeAlias = list[dict[VersionNumber, list[VersionInfo]]]


@final
class VersionNotFoundError(Exception):
    pass


class VersionDelta(Protocol):

    def days(self) -> int: ...


class SortedVersions(Protocol):

    def fetch(self) -> SortedVersionsList: ...


@final
@attrs.define(frozen=True)
class VersionsSortedByDate(SortedVersions):

    _package_name: str

    def fetch(self) -> SortedVersionsList:
        response = httpx.get('https://pypi.org/pypi/{0}/json'.format(self._package_name))
        response.raise_for_status()
        versions = list(response.json()['releases'].items())
        correct_versions = []
        for version_number, release_info in versions:
            with suppress(version.InvalidVersion):
                if not version.parse(version_number).pre:
                    correct_versions.append({
                        version_number: release_info,
                    })
        def _sort_key(release_dict: VersionInfo) -> datetime.date:
            times = []
            if not release_dict or not next(iter(list(release_dict.values()))):
                return datetime.date(1, 1, 1)
            times = [  # type: ignore[var-annotated]
                datetime.datetime.strptime(
                    dict_['upload_time'], '%Y-%m-%dT%H:%M:%S',
                ).astimezone(datetime.timezone.utc).date()
                for dict_ in chain.from_iterable(release_dict.values())  # type: ignore[arg-type]
            ]
            return next(iter(sorted(times)))
        return sorted(
            correct_versions,
            key=_sort_key,  # type: ignore[arg-type]
        )

@final
@attrs.define(frozen=True)
class VersionsSortedBySemver(SortedVersions):

    _artifactory_domain: str
    _package_name: str

    def fetch(self) -> SortedVersionsList:
        response = httpx.get(
            httpx.URL(self._artifactory_domain).join('pypi/{0}/json'.format(self._package_name)),
        )
        response.raise_for_status()
        versions = list(response.json()['releases'].items())
        correct_versions = []
        for version_number, release_info in versions:
            with suppress(version.InvalidVersion):
                v = version.parse(version_number)
                if not v.pre and not v.dev:
                    correct_versions.append({
                        version_number: release_info,
                    })
        return sorted(
            correct_versions,
            key=lambda release_dict: version.parse(next(iter(release_dict.keys()))),
        )


@final
@attrs.define(frozen=True)
class PypiVersionDelta(VersionDelta):

    _sorted_versions: SortedVersions
    _version: str

    def days(self) -> int:
        sorted_versions = self._sorted_versions.fetch()
        if not sorted_versions:
            return 0
        if next(iter(sorted_versions[-1].keys())) == self._version:
            return 0
        start = None
        next_available_release = None
        flag = False
        for release_info in sorted_versions:
            release_number = next(iter(release_info.keys()))
            if flag:
                next_available_release = release_number
                break
            if release_number == self._version:
                flag = True
        for item in sorted_versions:
            release = next(iter(item.keys()))
            item_values = list(chain.from_iterable(item.values()))
            if not item_values:
                continue
            upload_time = datetime.datetime.strptime(
                item_values[0]['upload_time'], '%Y-%m-%dT%H:%M:%S',
            ).astimezone(datetime.timezone.utc).date()
            if release == next_available_release:
                start = upload_time
        if not start:
            raise VersionNotFoundError
        return (datetime.datetime.now(tz=datetime.timezone.utc).date() - start).days



from pathlib import Path
from typing import Annotated

import typer
from rich import print
from rich.console import Console
from rich.progress import track
from rich.table import Table
from typing_extensions import TypeAlias

from deltaver.config import CliOrPyprojectConfig, Config, ConfigDict, PyprojectTomlConfig
from deltaver.parsed_requirements import ExcludedReqs, FileNotFoundSafeReqs, FreezedReqs, PoetryLockReqs
from deltaver.version_delta import PypiVersionDelta, VersionsSortedBySemver

app = typer.Typer()
PackageName: TypeAlias = str
PackageVersion: TypeAlias = str
PackageDelta: TypeAlias = str


def results(
    packages: list[tuple[PackageName, PackageVersion, PackageDelta]],
    sum_delta: int,
    max_delta: int,
    config: Config,
) -> tuple[int, float]:
    console = Console()
    table = Table(show_header=True, header_style='bold magenta')
    table.add_column('Package')
    table.add_column('Version')
    table.add_column('Delta (days)')
    for package, version, delta in packages:
        table.add_row(package, version, delta)
    if len(packages) > 0:
        console.print(table)
        average_delta = '{0:.2f}'.format(sum_delta / len(packages))
    else:
        average_delta = '0'
    print('Max delta: {0}'.format(max_delta))
    print('Average delta: {0}'.format(average_delta))
    if config.value_of('fail_on_avg') > -1 and float(average_delta) >= config.value_of('fail_on_avg'):
        print('\n[red]Error: average delta greater than available[/red]')
        raise typer.Exit(code=1)
    if config.value_of('fail_on_max') > -1 and max_delta >= config.value_of('fail_on_max'):
        print('\n[red]Error: max delta greater than available[/red]')
        raise typer.Exit(code=1)


@app.command()
def main(  # noqa: PLR0913
    path_to_requirements_file: str,
    file_format: Annotated[str, typer.Option('--format')] = 'freezed',
    fail_on_average: Annotated[int, typer.Option('--fail-on-avg')] = -1,
    fail_on_max: Annotated[int, typer.Option('--fail-on-max')] = -1,
    artifactory_domain: Annotated[str, typer.Option('--artifactory-domain')] = 'https://pypi.org',
    exclude_deps: Annotated[list[str], typer.Option('--exclude')] = [],  # noqa: B006
) -> None:
    res = 0
    max_delta = 0
    packages = []
    config = CliOrPyprojectConfig(
        PyprojectTomlConfig(Path('pyproject.toml')),
        ConfigDict({
            'file_format': file_format,
            'fail_on_avg': fail_on_average,
            'fail_on_max': fail_on_max,
            'artifactory_domain': artifactory_domain,
            'excluded': exclude_deps,
        }),
    )
    reqs_obj_ctor = {
        'freezed': FreezedReqs,
        'lock': PoetryLockReqs,
    }[config.value_of('file_format')]
    dependencies = FileNotFoundSafeReqs(
        ExcludedReqs(
            reqs_obj_ctor(Path(path_to_requirements_file)),
            config,
        ),
    ).reqs()
    for package, version in track(dependencies, description='Scanning...'):
        delta = PypiVersionDelta(
            VersionsSortedBySemver(
                config.value_of('artifactory_domain'),
                package,
            ),
            version,
        ).days()
        if delta > 0:
            packages.append(
                (package, version, str(delta)),
            )
        res += delta
        max_delta = max(max_delta, delta)
    packages = sorted(packages, key=lambda x: int(x[2]), reverse=True)
    results(packages, res, max_delta, config)


if __name__ == '__main__':
    app()



import enum
from pathlib import Path
from typing import Protocol, TypedDict, TypeVar, final

import attrs
import toml

ConfigValueT = TypeVar('ConfigValueT')


@final
class PyprojectNotContainDeltaverConfigError(Exception):
    pass


class Formats(enum.Enum):

    freezed = 'freezed'
    lock = 'lock'


class ConfigDict(TypedDict):

    file_format: Formats
    fail_on_avg: int
    fail_on_max: int
    artifactory_domain: str
    excluded: list[str]


class Config(Protocol):

    def value_of(self, value: str) -> ConfigValueT: ...


@final
@attrs.define(frozen=True)
class PyprojectTomlConfig(Config):

    _path: Path

    def value_of(self, value: str) -> ConfigValueT:
        try:
            deltaver_config = toml.loads(self._path.read_text())['tool']['deltaver']
        except KeyError as err:
            raise PyprojectNotContainDeltaverConfigError from err
        return deltaver_config[value]


@final
@attrs.define(frozen=True)
class CliOrPyprojectConfig(Config):

    _pyproject_config: Config
    _cli_configs: ConfigDict

    def value_of(self, value: str) -> ConfigValueT:
        try:
            return self._pyproject_config.value_of(value)
        except (KeyError, PyprojectNotContainDeltaverConfigError):
            return self._cli_configs[value]



import re
from pathlib import Path
from typing import Protocol, final

import attrs
import toml
import typer
from rich import print

from deltaver.config import Config


class ParsedReqs(Protocol):

    def reqs(self) -> list[tuple[str, str]]: ...


@final
@attrs.define(frozen=True)
class FreezedReqs(ParsedReqs):

    _path: Path

    def reqs(self) -> list[tuple[str, str]]:
        res = []
        lines = self._path.read_text().strip().splitlines()
        expected_splitted_line_len = 2
        for line in lines:
            splitted_line = line.split(';')[0].split('==')
            if len(splitted_line) != expected_splitted_line_len:
                continue
            package, version = splitted_line
            package = re.sub(r'\[.*?\]', '', package)
            res.append((package, version.strip()))
        return res


@final
@attrs.define(frozen=True)
class ExcludedReqs(ParsedReqs):

    _origin: ParsedReqs
    _config: Config

    def reqs(self) -> list[tuple[str, str]]:
        excluded_packages_set = {package.lower() for package in self._config.value_of('excluded')}
        return [
            item
            for item in self._origin.reqs()
            if item[0].lower() not in excluded_packages_set
        ]


@final
@attrs.define(frozen=True)
class PoetryLockReqs(ParsedReqs):

    _path: Path

    def reqs(self) -> list[tuple[str, str]]:
        data = toml.loads(self._path.read_text())
        return [
            (dependency['name'], dependency['version'])
            for dependency in data['package']
        ]


@final
@attrs.define(frozen=True)
class FileNotFoundSafeReqs(ParsedReqs):

    _origin: ParsedReqs

    def reqs(self) -> list[tuple[str, str]]:
        try:
            return self._origin.reqs()
        except FileNotFoundError:
            print('Requirements file not found')
            typer.Exit(1)
